﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

using aDictionaryExtension;

namespace aMultipleInheritanceHelpler
{
	class zTypeName
	{		
		/// <summary>
		/// C#的に正しい型名を返す。isOutは引数の型が参照型である可能性がある場合にrefかoutか判断できるようにする。
		/// 通常の型では必要ない。
		/// </summary>
		public static string TypeName(Type type, bool isOut, Dictionary<string, string> typeMap)
		{
			//ネストされたクラスでは親を遡る。
			//しかし、たとえばList<T>のTはList<T>のインナークラスと判定されるようだ。だからそこでは遡らない。
			if (type.IsGenericParameter == false && type.IsNested)
			{
				var declares = Declares(type);

				var parentresults = declares.Take(declares.Length - 1).Select(t => ParentTypeName(t, typeMap));

				var lastResult = LastTypeName(declares.Last(), isOut, typeMap);

				return ToTypeName(lastResult, parentresults.ToArray());
			}
			var r = LastTypeName(type, isOut, typeMap);
			return ToTypeName(r, new Result[0]);
		}

		/// <summary>
		/// lastは全ての型引数の正確な名前を持ち
		/// parentsは型引数の数を持っている。
		/// 親から順に型引数を持って行く
		/// </summary>
		static string ToTypeName(LastResult last, Result[] parents)
		{
			var first = ToStr(last.OutOrRef);

			//型引数つきの名前に変換
			Func<string, string[], string> toTypeName = (name, array) =>
			{
				if(array.Length != 0)
				{
					return name + "<" + String.Join(", ", array) + ">";
				}
				return name;
			};

			int index = 0;

			var parentNames = parents.Select(p =>
			{
				//必要な分だけ型引数をもらう
				var array = last.GenericArgTypeNames
					.Skip(index).Take(p.NumGenericArgs).ToArray();

				index += p.NumGenericArgs;

				return toTypeName(p.Name, array);
			}).ToArray();

			var lastGenericArgs = last.GenericArgTypeNames.Skip(index).ToArray();

			var lastName = toTypeName(last.Name,
				last.GenericArgTypeNames.Skip(index).ToArray());

			return ToStr(last.OutOrRef) + 
				parentNames.Aggregate(new StringBuilder(), (sb, name) => sb.Append(name + ".")) + lastName + last.ArrayStr;
		}

		static string ToStr(RefType r)
		{
			switch (r)
			{
				case RefType.None: return "";
				case RefType.Out: return "out ";
				case RefType.Ref: return "ref ";
				default: throw new Exception();
			}
		}

		static Regex typeCount = new Regex(@"(\`([0-9]+))");
		static Regex array_Kakko = new Regex(@"([\[\]\,]+)");

		/// <summary>
		/// ネストしたクラスの親の方のクラスの情報を返す。型引数の正確な情報を持っていない。
		/// </summary>
		static Result ParentTypeName(Type type, Dictionary<string, string> typeMap)
		{
			var name = type.Name;
			var index = name.IndexOf('`');
			int numGenericArgs = 0;

			var match = typeCount.Match(name);

			//`1みたいな型引数の数を取り出す。
			if (match.Success)
			{
				numGenericArgs = int.Parse(match.Groups[2].Value);
				var g = match.Groups[1];
				name = name.Remove(g.Index, g.Length);
			}

			//配列の内部クラスとかあると困るな。ないよな。
			name = typeMap.GetValueN(name) ?? name;

			return new Result { Name = name, NumGenericArgs = numGenericArgs };
		}

		/// <summary>
		/// ネストしたクラスなら一番下、ネストしてないなら通常のクラス。
		/// パラメータのref,outが入る可能性があり、また確定した型引数もここからしか取れない。
		/// </summary>
		static LastResult LastTypeName(Type type, bool isOut, Dictionary<string, string> typeMap)
		{
			var name = type.Name;
			int numGenericArgs = 0;
			RefType outOrRef = RefType.None;

			var match = typeCount.Match(name);

			//`1みたいな型引数の数を取り出す。
			if (match.Success)
			{
				numGenericArgs = int.Parse(match.Groups[2].Value);
				var g = match.Groups[1];
				name = name.Remove(g.Index, g.Length);
			}

			if (name.EndsWith("&"))
			{
				name = name.Remove(name.Length - 1, 1);

				outOrRef = isOut ? RefType.Out : RefType.Ref;
			}

			var arrayMatch = array_Kakko.Match(name);
			string arrayStr = "";
			if (arrayMatch.Success)
			{
				arrayStr = name.Substring(arrayMatch.Index, arrayMatch.Length);
				//[] [][] [,,] みたいな配列の記号を切る
				name = name.Remove(arrayMatch.Index, arrayMatch.Length);
			}

			name = typeMap.GetValueN(name) ?? name;

			var array = type.GetGenericArguments();

			var args = array.Select(t => TypeName(t, false, typeMap)).ToArray();

			return new LastResult
			{
				Name = name,
				OutOrRef = outOrRef,
				NumGenericArgs = numGenericArgs,
				ArrayStr = arrayStr,
				GenericArgTypeNames = args
			};
		}

		/// <summary>
		/// 入れ子になった型を頭から順に返す
		/// </summary>
		static Type[] Declares(Type type)
		{
			return InnerDeclares(type).Reverse().ToArray();
		}

		static IEnumerable<Type> InnerDeclares(Type type)
		{
			yield return type;

			while (type.IsNested)
			{
				yield return type.DeclaringType;

				type = type.DeclaringType;
			}
		}

		enum RefType { None, Out, Ref }

		struct LastResult
		{
			public RefType OutOrRef;
			public string[] GenericArgTypeNames;
			public int NumGenericArgs;
			public string Name;
			/// <summary>
			/// []とか[][]とか[,]とか
			/// </summary>
			public string ArrayStr;
		}

		/// <summary>
		/// ネストした型から帰る情報。
		/// </summary>
		struct Result
		{
			public string Name;
			public int NumGenericArgs;
		}
	}
}
