﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using System.Reflection;

using aDictionaryExtension;

namespace aSourceRewriter
{
	/// <summary>
	/// ソースファイルを見つけて書き換える
	/// </summary>
	public class zSourceRewriter
	{
		public Dictionary<string, string[]> Dict;

		public string SolutionFolderPath { get; set; }

		/// <summary>
		/// 書き込む先を指定しなかった場合ここになる
		/// </summary>
		public string DefaultFileRelPath { get; set; }

		public string DefaultFilePath { get { return Path.Combine(SolutionFolderPath, DefaultFileRelPath); } }

		/// <summary>
		/// デフォルトコンストラクタ。初期化は自力で。
		/// </summary>
		public zSourceRewriter() { }

		/// <summary>
		/// ソースファイルを見つけて書き換える
		/// </summary>
		/// <param name="solutionFolderName">ソリューションのフォルダ名。パスではない</param>
		/// <param name="defaultFileRelPath">ソリューションフォルダからの相対パス。書き込む先を指定しなかった場合ここになる</param>
		public zSourceRewriter(string solutionFolderName, string defaultFileRelPath)
		{
			SolutionFolderPath = SearchSolutionFolderN(solutionFolderName, Assembly.GetExecutingAssembly().Location);
			if (SolutionFolderPath == null)
				throw new Exception("solutionFolderがexeからさかのぼっていったところに見つかりません");

			DefaultFileRelPath = defaultFileRelPath;

			Dict = CreateDictionary(SolutionFolderPath);
		}

		public static Dictionary<string, string[]> CreateDictionary(string solutionFolderPath)
		{
			return GetFiles(solutionFolderPath)
				.GroupBy(path => Path.GetFileName(path))
				.Select(g => new { Key = g.Key, Pathes = g.ToArray() })
				.ToDictionary(item => item.Key, item => item.Pathes);
		}

		public static string SearchSolutionFolderN(string solName, string exeFilePath)
		{
			var path = exeFilePath;
			while (path != null)
			{
				if (Path.GetFileName(path) == solName)
					return path;

				path = Path.GetDirectoryName(path);
			}
			return null;
		}

		public static IEnumerable<string> GetFiles(string path)
		{
			var dirs = Directory.GetDirectories(path);

			foreach (var dir in dirs)
			{
				foreach (var file in GetFiles(dir))
				{
					yield return file;
				}
			}

			var files = Directory.GetFiles(path)
				.Where(file => Path.GetExtension(file) == ".cs");

			foreach (var file in files)
			{
				yield return file;
			}
		}

		/// <summary>
		/// パスでも相対パスでもなくファイル名(***.cs)を入れる。
		/// </summary>
		public StreamWriter GetWriter(string fileName)
		{
			if (Dict.ContainsKey(fileName) == false)
				throw new FileNotFoundException(fileName + " が見つかりません");

			var array = Dict[fileName];

			if (array.Length > 1)
				throw new ApplicationException(fileName + " が重複しています\r\n\r\n" +
					String.Join("\r\n", array));

			var filePath = array[0];

			return PathToWriter(filePath);
		}

		StreamWriter PathToWriter(string path)
		{
			if (File.Exists(path))
				return new StreamWriter(new FileStream(path, FileMode.Truncate));

			throw new FileNotFoundException(path + "が見つかりません");
		}

		public StreamWriter GetWriter()
		{
			return PathToWriter(DefaultFilePath);
		}

		/// <summary>
		/// ファイル名(***.cs)をいれる。
		/// </summary>
		public void Write(string fileName, string sourceCode)
		{
			using (var writer = GetWriter(fileName))
			{
				writer.Write(sourceCode);
			}
		}

		/// <summary>
		/// デフォルトのファイルに書き込む
		/// </summary>
		public void Write(string sourceCode)
		{
			using (var writer = GetWriter())
			{
				writer.Write(sourceCode);
			}
		}

		/// <summary>
		/// パスじゃなくファイル名を入れる。
		/// </summary>
		public string[] GetSourceCode(string fileName)
		{
			var array = Dict[fileName];

			if (array.Length > 1)
				throw new Exception(fileName + "が重複しています");

			if (array.Length == 0)
				throw new FileNotFoundException(fileName = "が見つかりません");

			return File.ReadAllLines(array[0]);
		}
	}
}
